# Copyright (C) 2022--2023 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""System-wide facilities.
"""

from __future__ import annotations

import os
from pathlib import Path
import subprocess
import sys

from loguru import logger

from baldaquin._version import VERSION as __version__, TAG_DATE


def start_message() -> None:
    """Print the start message.
    """
    msg = f"""
    This is baldaquin version {__version__} ({TAG_DATE}).

    Copyright (C) 2022--2024, the baldaquin team.

    baldaquin comes with ABSOLUTELY NO WARRANTY.
    This is free software, and you are welcome to redistribute it under certain
    conditions. See the LICENSE file for details.

    Visit https://github.com/lucabaldini/baldaquin for more information.
    """
    print(msg)


# Logger setup.
DEFAULT_LOGURU_HANDLER = dict(sink=sys.stderr, colorize=True,
                              format=">>> <level>[{level}] {message}</level>")


def config_logger(file_path: str = None, extra=None):
    """Configure the loguru logger.
    """
    handlers = [DEFAULT_LOGURU_HANDLER]
    if file_path is not None:
        handlers.append(dict(sink=file_path, enqueue=True, serialize=True))
    logger.configure(handlers=handlers, levels=None, extra=extra)


config_logger()

# Basic package structure.
BALDAQUIN_ROOT = Path(__file__).parent
BALDAQUIN_BASE = BALDAQUIN_ROOT.parent
BALDAQUIN_GRAPHICS = BALDAQUIN_ROOT / 'graphics'
BALDAQUIN_ICONS = BALDAQUIN_GRAPHICS / 'icons'
BALDAQUIN_SKINS = BALDAQUIN_GRAPHICS / 'skins'
BALDAQUIN_DOCS = BALDAQUIN_BASE / 'docs'
BALDAQUIN_DOCS_STATIC = BALDAQUIN_DOCS / '_static'
BALDAQUIN_TESTS = BALDAQUIN_BASE / 'tests'
BALDAQUIN_TEST_DATA = BALDAQUIN_TESTS / 'data'


# Version information.
BALDAQUIN_VERSION_FILE_PATH = BALDAQUIN_ROOT / '_version.py'


# pyproject.toml file.
BALDAQUIN_TOML_FILE_PATH = BALDAQUIN_BASE / 'pyproject.toml'


# Release notes file.
BALDAQUIN_RELEASE_NOTES_PATH = BALDAQUIN_DOCS / 'release_notes.rst'


# Default character encoding.
DEFAULT_CHARACTER_ENCODING = 'utf-8'


def execute_shell_command(args):
    """Execute a shell command.
    """
    logger.info(f'About to execute "{" ".join(args)}"...')
    return subprocess.run(args, check=True)


def _create_folder(folder_path: Path) -> None:
    """Create a given folder if it does not exist.

    This is a small utility function to ensure that the relevant directories
    exist when needed at runtime.

    Arguments
    ---------
    folder_path : Path instance
        The path to the target folder.
    """
    if not folder_path.exists():
        logger.info(f'Creating folder {folder_path}...')
        Path.mkdir(folder_path, parents=True)


# The path to the base folder for the output data defaults to ~/baldaquindata,
# but can be changed via the $BALDAQUIN_DATA environmental variable.
try:
    BALDAQUIN_DATA = Path(os.environ['BALDAQUIN_DATA'])
except KeyError:
    BALDAQUIN_DATA = Path.home() / 'baldaquindata'
_create_folder(BALDAQUIN_DATA)


# We're doing a similar thing for our scratch space.
try:
    BALDAQUIN_SCRATCH = Path(os.environ['BALDAQUIN_SCRATCH'])
except KeyError:
    BALDAQUIN_SCRATCH = BALDAQUIN_DATA / 'scratch'
_create_folder(BALDAQUIN_SCRATCH)


# On the other hand all the configuration files live in (subdirectories of) ~/.baldaquin
BALDAQUIN_CONFIG = Path.home() / '.baldaquin'
_create_folder(BALDAQUIN_CONFIG)


def config_folder_path(project_name: str) -> Path:
    """Return the path to the configuration folder for a given project.

    Arguments
    ---------
    project_name : str
        The name of the project.
    """
    return BALDAQUIN_CONFIG / project_name


def data_folder_path(project_name: str) -> Path:
    """Return the path to the data folder for a given project.

    Arguments
    ---------
    project_name : str
        The name of the project.
    """
    return BALDAQUIN_DATA / project_name


def setup_project(project_name: str) -> tuple[Path, Path]:
    """Setup the folder structure for a given project.

    This is essentially creating a folder for the configuration files and
    a folder for the data files, if they do not exist already, and returns
    the path to the two (in this order---first config and then data).

    Arguments
    ---------
    project_name : str
        The name of the project.
    """
    config_folder = config_folder_path(project_name)
    app_config_folder = config_folder / 'apps'
    data_folder = data_folder_path(project_name)
    folder_list = (config_folder, app_config_folder, data_folder)
    for folder_path in folder_list:
        _create_folder(folder_path)
    return folder_list



# Copyright (C) 2022 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Test suite for runctrl.py
"""

import pytest

from baldaquin.runctrl import FiniteStateMachineBase, InvalidFsmTransitionError


def test_finite_state_machine():
    """This creates a mock FSM where all the relevant hooks are no-op, and tests
    for the proper states after transitions, as well as the exceptions signaling
    the invalid ones.
    """
    # Create a virtual FSM.
    fsm = FiniteStateMachineBase()
    # Override all the necessary hooks to that we can actually have transitions.
    fsm.setup = lambda: True
    fsm.teardown = lambda: True
    fsm.start_run = lambda: True
    fsm.stop_run = lambda: True
    fsm.pause = lambda: True
    fsm.stop = lambda: True
    fsm.resume = lambda: True
    # Start in the RESET state.
    assert fsm.is_reset()
    with pytest.raises(InvalidFsmTransitionError):
        fsm.set_running()
    with pytest.raises(InvalidFsmTransitionError):
        fsm.set_reset()
    with pytest.raises(InvalidFsmTransitionError):
        fsm.set_paused()
    # Transition to STOPPED.
    fsm.set_stopped()
    assert fsm.is_stopped()
    with pytest.raises(InvalidFsmTransitionError):
        fsm.set_stopped()
    with pytest.raises(InvalidFsmTransitionError):
        fsm.set_paused()
    # Transition to RUNNING.
    fsm.set_running()
    assert fsm.is_running()
    with pytest.raises(InvalidFsmTransitionError):
        fsm.set_reset()
    # Transition to PAUSED.
    fsm.set_paused()
    assert fsm.is_paused()
    with pytest.raises(InvalidFsmTransitionError):
        fsm.set_reset()
    # And now back to back and forth from/to STOPPED and to the RESET state...
    fsm.set_running()
    fsm.set_stopped()
    fsm.set_running()
    fsm.set_paused()
    fsm.set_stopped()
    fsm.set_reset()
    assert fsm.is_reset()



# Copyright (C) 2025 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""baldaquin command-line interface.
"""

from __future__ import annotations

import argparse
import importlib
import importlib.util
import sys

from baldaquin import logger, BALDAQUIN_ROOT, BALDAQUIN_DATA
from baldaquin import arduino_
from baldaquin import serial_


# List of defaults projects shipped with the package.
_DEFAULT_PROJECTS = ('plasduino', 'silly')


class _Formatter(argparse.RawDescriptionHelpFormatter, argparse.ArgumentDefaultsHelpFormatter):

    """Do nothing class combining our favorite formatting for the
    command-line options, i.e., the newlines in the descriptions are
    preserved and, at the same time, the argument defaults are printed
    out when the --help options is passed.

    The inspiration for this is coming from one of the comments in
    https://stackoverflow.com/questions/3853722
    """


class MainArgumentParser(argparse.ArgumentParser):

    """Application-wide argument parser.
    """

    _DESCRIPTION = None
    _EPILOG = None
    _FORMATTER_CLASS = _Formatter

    def __init__(self) -> None:
        """Overloaded method.
        """
        super().__init__(description=self._DESCRIPTION, epilog=self._EPILOG,
                         formatter_class=self._FORMATTER_CLASS)
        subparsers = self.add_subparsers(required=True, help='sub-command help')
        # See https://stackoverflow.com/questions/8757338/
        subparsers._parser_class = argparse.ArgumentParser

        # Start a given application.
        start_app = subparsers.add_parser('start-app',
            help='start a baldaquin application',  # noqa: E128
            formatter_class=self._FORMATTER_CLASS)  # noqa: E128
        start_app.add_argument('app_name', choices=self.list_apps())
        start_app.set_defaults(func=self.start_app)

        # Simply list the COM ports.
        list_com_ports = subparsers.add_parser('list-com-ports',
            help='list the available COM ports',  # noqa: E128
            formatter_class=self._FORMATTER_CLASS)  # noqa: E128
        list_com_ports.set_defaults(func=serial_.list_com_ports)

        # Arduino autodetect.
        arduino_autodetect = subparsers.add_parser('arduino-autodetect',
            help='autodetect arduino boards attached to the COM ports',  # noqa: E128
            formatter_class=self._FORMATTER_CLASS)  # noqa: E128
        arduino_autodetect.set_defaults(func=arduino_.autodetect_arduino_boards)

        # Arduino upload.
        arduino_compile = subparsers.add_parser('arduino-compile',
            help='compile a sketch for a given arduino board',  # noqa: E128
            formatter_class=self._FORMATTER_CLASS)  # noqa: E128
        arduino_compile.add_argument('file_path',
            help='the path to the sketch source file')  # noqa: E128
        arduino_compile.add_argument('--output_dir', default=BALDAQUIN_DATA)
        arduino_compile.add_argument('--board_designator', default='uno')
        arduino_compile.set_defaults(func=arduino_.compile_sketch)

        # Arduino upload.
        arduino_upload = subparsers.add_parser('arduino-upload',
            help='upload a sketch to an arduino board',  # noqa: E128
            formatter_class=self._FORMATTER_CLASS)  # noqa: E128
        arduino_upload.add_argument('file_path',
            help='the path to the compiled sketch file')  # noqa: E128
        arduino_upload.add_argument('--board', default=arduino_.UNO)
        arduino_upload.set_defaults(func=arduino_.upload_sketch)

    @staticmethod
    def list_apps() -> list[str]:
        """List all the available applications.

        This is looping over all the registered project folders and searching
        for Python modules. Note that ``pathlib.Path.iterdir()`` returns a list
        in no particular order, and we do sort the list in place for convenience.
        """
        apps = []
        for project in _DEFAULT_PROJECTS:
            folder_path = BALDAQUIN_ROOT / project / 'apps'
            apps += [_path.stem for _path in folder_path.iterdir() if _path.suffix == '.py']
        apps.sort()
        return apps

    @staticmethod
    def start_app(app_name: str) -> None:
        """Start a given application.

        Arguments
        ---------
        app_name : str
            The application name.
        """
        logger.info('Starting application...')
        # Loop over the project folders and search for a Python module matching the
        # target application name.
        for project in _DEFAULT_PROJECTS:
            folder_path = BALDAQUIN_ROOT / project / 'apps'
            file_path = folder_path / f'{app_name}.py'
            if file_path.exists():
                logger.debug(f'Adding {file_path} to sys.path...')
                sys.path.append(str(folder_path))
                break
        # At this point we do want to import the module without generating the
        # bytecode---we cannot assume the thing is in a folder with write priviledges.
        sys.dont_write_bytecode = True
        module = importlib.import_module(app_name)
        sys.dont_write_bytecode = False
        # And, finally, we call the application entry point.
        module.main()

    def run_command(self) -> None:
        """Run the actual command tied to the specific options.
        """
        kwargs = vars(self.parse_args())
        command = kwargs.pop('func')
        command(**kwargs)


def main() -> None:
    """Main entry point.
    """
    MainArgumentParser().run_command()


if __name__ == '__main__':
    main()



# Copyright (C) 2024 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Strip-chart facilities.
"""

import collections

from baldaquin.plt_ import plt, setup_axes


class SlidingStripChart:

    """Class describing a strip chart, that is, a scatter plot where the number of
    points is limited to a maximum, so that the thing acts essentially as a sliding
    window, typically in time.

    This is mainly meant to represent the time history of a signal over a reasonable
    span---a long-term acquisition might go on for weeks, and it would not make sense
    to try and plot on the screen millions of points, but the last segment of the
    acquisition is the most important part when we want to monitor what is happening.
    """

    # pylint: disable=invalid-name

    def __init__(self, max_length: int = None, label: str = '', xoffset: float = 0.,
                 xlabel: str = None, ylabel: str = None) -> None:
        """Constructor.
        """
        self.label = label
        self.xoffset = xoffset
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.x = collections.deque(maxlen=max_length)
        self.y = collections.deque(maxlen=max_length)

    def reset(self, max_length: int = None) -> None:
        """Reset the strip chart.
        """
        self.x = collections.deque(maxlen=max_length)
        self.y = collections.deque(maxlen=max_length)

    def add_point(self, x: float, y: float) -> None:
        """Append a data point to the strip chart.
        """
        self.x.append(x + self.xoffset)
        self.y.append(y)

    def plot(self, axes=None, **kwargs) -> None:
        """Plot the strip chart.
        """
        if axes is None:
            axes = plt.gca()
        axes.plot(self.x, self.y, label=self.label, **kwargs)
        setup_axes(axes, xlabel=self.xlabel, ylabel=self.ylabel, grids=True)



# Copyright (C) 2022 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Test suite for config.py
"""

import os

from baldaquin import BALDAQUIN_DATA
from baldaquin.__qt__ import exec_qapp
from baldaquin.config import ConfigurationParameter, SampleConfiguration
from baldaquin.gui import ConfigurationWidget, bootstrap_qapplication


def _test_base_match(type_name, value, **constraints):
    """Base test function where we expect the parameter to match the input value.
    """
    p = ConfigurationParameter('parameter', type_name, value, '', **constraints)
    assert p.value == value


def _test_base_mismatch(type_name, value, **constraints):
    """Base test function where we expect the parameter to match the input value.
    """
    p = ConfigurationParameter('parameter', type_name, value, '', **constraints)
    assert p.value is None


def test_print_parameters():
    """Print some parameters.
    """
    print(ConfigurationParameter('timeout', 'float', 10., 'timeout'))
    print(ConfigurationParameter('timeout', 'float', 10., 'timeout', 's'))
    print(ConfigurationParameter('timeout', 'float', 10., 'timeout', 's', '.6f'))
    print(ConfigurationParameter('timeout', 'float', 10., 'timeout', 's', '.6f', min=0.))


def test_parameter_bool():
    """Test possible setting for int parameters.
    """
    for value in (True, False):
        _test_base_match('bool', value)
    for value in (1, 1., 'test'):
        _test_base_mismatch('bool', value)


def test_parameter_int(value=100):
    """Test possible setting for int parameters.
    """
    _test_base_match('int', 100)
    _test_base_match('int', 100, min=0, max=1000)
    _test_base_mismatch('int', 100, min=200)
    _test_base_mismatch('int', 100, max=0)
    _test_base_match('int', 100, choices=(99, 100, 101))
    _test_base_mismatch('int', 100, choices=(99, 101))
    _test_base_match('int', 100, step=10)
    _test_base_mismatch('int', 100, step=13)
    _test_base_match('int', 100, min=87, step=13)


def test_parameter_float(value=1.):
    """Test possible setting for int parameters.
    """
    _test_base_match('float', 1.)
    _test_base_match('float', 1., min=0., max=10.)
    _test_base_mismatch('float', 1., min=10.)
    _test_base_mismatch('float', 1., max=0.)


def test_parameter_str():
    """Test possible setting for int parameters.
    """
    _test_base_match('str', 'howdy?')
    _test_base_match('str', 'eggs', choices=('eggs', 'cheese'))
    _test_base_mismatch('str', 'ham', choices=('eggs', 'cheese'))
    for value in (True, 1, 1.):
        _test_base_mismatch('str', value)


def test_configuration():
    """Test an actual, fully-fledged configuration.
    """
    config = SampleConfiguration()
    print(config)
    file_path = os.path.join(BALDAQUIN_DATA, 'test_config.json')
    config.save(file_path)
    config.update(file_path)
    assert config.value('port') == 20004
    config.update_value('port', 3)
    assert config.value('port') == 20004
    config.update_value('port', 20003)
    config.save(file_path)
    config = SampleConfiguration()
    config.update(file_path)
    assert config.value('port') == 20003


def _test_configuration_display(port=9999):
    """Test the widget displaying configuration objects.
    """
    config = SampleConfiguration()
    print(config)
    assert config.value('port') != port
    qapp = bootstrap_qapplication()
    widget = ConfigurationWidget(config)
    widget.set_value('port', port)
    config = widget.current_configuration()
    print(config)
    assert config.value('port') == port
    return qapp, widget


if __name__ == '__main__':
    qapp, widget = _test_configuration_display()
    widget.show()
    exec_qapp(qapp)



# Copyright (C) 2022 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Configuration facilities.
"""

import enum
import json
import os
from typing import Any

from baldaquin import logger, DEFAULT_CHARACTER_ENCODING


class ParameterValidationError(enum.IntEnum):

    """Enum class for the possible errors occurring while checking the input
    parameter values.
    """

    PARAMETER_VALID = 0
    INVALID_TYPE = enum.auto()
    NUMBER_TOO_SMALL = enum.auto()
    NUMBER_TOO_LARGE = enum.auto()
    INVALID_CHOICE = enum.auto()
    INVALID_STEP = enum.auto()
    GENERIC_ERROR = enum.auto()


class ConfigurationParameter:

    """Class representing a configuration parameter.

    This is a simple attempt at putting in place a generic configuration mechanism
    where we have some control on the values we are passing along.

    A configuration parameter is fully specified by its name, type and value, and
    When setting the latter, we make sure that the its type matches.
    Additional, we can specify simple conditions on the parameters that are
    then enforced at runtime.

    Arguments
    ---------
    name : str
        The parameter name.

    type_name : str
        The name of the parameter type.

    value : anything
        The parameter value.

    intent : str
        The intent of the parameter, acting as a comment in the corresponding
        configuration file.

    units : str, optional
        The units for the configuration parameter.

    fmt : str, optional
        An optional format string for the preferred rendering the parameter value.

    constraints : dict, optional
        A dictionary containing optional specifications on the parameter value.
    """

    VALID_CONSTRAINTS = {
        'int': ('choices', 'step', 'min', 'max'),
        'float': ('min', 'max'),
        'str': ('choices',)
    }

    def __init__(self, name: str, type_name: str, value: Any, intent: str,
                 units: str = None, fmt: None = None, **constraints) -> None:
        """Constructor.
        """
        self.name = name
        self.type_name = type_name
        self.value = None
        self.intent = intent
        self.units = units
        self.fmt = fmt
        for key in tuple(constraints):
            if key not in self.VALID_CONSTRAINTS.get(self.type_name, ()):
                logger.warning(f'Invalid spec ({key}) for {self.name} ({self.type_name})...')
                constraints.pop(key)
        self.constraints = constraints
        self.set_value(value)

    def not_set(self) -> bool:
        """Return true if the parameter value is not set.
        """
        return self.value is None

    def _validation_error(self, value: Any,
                          error_code: ParameterValidationError) -> ParameterValidationError:
        """Utility function to log a parameter error (and forward the error code).
        """
        logger.error(f'Value {value} invalid for {self.name} {self.constraints}: {error_code.name}')
        logger.error('Parameter value will not be set')
        return error_code

    def _check_range(self, value: Any) -> ParameterValidationError:
        """Generic function to check that a given value is within a specified range.

        This is used for validating int and float parameters.
        """
        if 'min' in self.constraints and value < self.constraints['min']:
            return self._validation_error(value, ParameterValidationError.NUMBER_TOO_SMALL)
        if 'max' in self.constraints and value > self.constraints['max']:
            return self._validation_error(value, ParameterValidationError.NUMBER_TOO_LARGE)
        return ParameterValidationError.PARAMETER_VALID

    def _check_choice(self, value: Any) -> ParameterValidationError:
        """Generic function to check that a parameter value is within the
        allowed choices.
        """
        if 'choices' in self.constraints and value not in self.constraints['choices']:
            return self._validation_error(value, ParameterValidationError.INVALID_CHOICE)
        return ParameterValidationError.PARAMETER_VALID

    def _check_step(self, value: int) -> ParameterValidationError:
        """Generic function to check the step size for an integer.
        """
        delta = value - self.constraints.get('min', 0)
        if 'step' in self.constraints and delta % self.constraints['step'] != 0:
            return self._validation_error(value, ParameterValidationError.INVALID_STEP)
        return ParameterValidationError.PARAMETER_VALID

    def _check_int(self, value: int) -> ParameterValidationError:
        """Validate an integer parameter value.

        Note we check the choice specification first, and all the others after that
        (this is relevant as, if you provide inconsistent conditions the order
        becomes relevant).
        """
        for check in (self._check_choice, self._check_range, self._check_step):
            error_code = check(value)
            if error_code != ParameterValidationError.PARAMETER_VALID:
                return error_code
        return ParameterValidationError.PARAMETER_VALID

    def _check_float(self, value: float) -> ParameterValidationError:
        """Validate a floating-point parameter value.
        """
        return self._check_range(value)

    def _check_str(self, value: str) -> ParameterValidationError:
        """Validate a string parameter value.
        """
        return self._check_choice(value)

    def set_value(self, value: Any) -> ParameterValidationError:
        """Set the paramater value.
        """
        # Make sure that type(value) matches the expectations.
        if value.__class__.__name__ != self.type_name:
            return self._validation_error(value, ParameterValidationError.INVALID_TYPE)
        # If a validator is defined for the specific parameter type, apply it.
        func_name = f'_check_{self.type_name}'
        if hasattr(self, func_name):
            error_code = getattr(self, func_name)(value)
            if error_code:
                return error_code
        # And if we made it all the way to this point we're good to go :-)
        self.value = value
        return ParameterValidationError.PARAMETER_VALID

    def __str__(self):
        """String formatting.
        """
        text = f'{self.name:.<20}: {self.value:{self.fmt if self.fmt else ""}}'
        if self.units is not None:
            text = f'{text} {self.units}'
        if len(self.constraints):
            text = f'{text} {self.constraints}'
        return text


class ConfigurationBase(dict):

    """Base class for configuration data structures.

    The basic idea, here, is that specific configuration classes simply override
    the ``TITLE`` and ``PARAMETER_SPECS`` class members. ``PARAMETER_SPECS``, particulalry,
    encodes the name, types and default values for all the configuration parameters,
    as well optional help strings and parameter constraints.

    Configuration objects provide file I/O through the JSON protocol. One
    important notion, here, is that configuration objects are always created
    in place with all the parameters set to their default values, and then updated
    from a configuration file. This ensures that the configuration is always
    valid, and provides an effective mechanism to be robust against updates of
    the configuration structure.
    """

    TITLE = 'Configuration'
    PARAMETER_SPECS = ()

    def __init__(self) -> None:
        """Constructor.
        """
        super().__init__()
        for *args, constraints in self.PARAMETER_SPECS:
            self.add_parameter(*args, **constraints)

    def add_parameter(self, *args, **kwargs) -> None:
        """Add a new parameter to the configuration.
        """
        parameter = ConfigurationParameter(*args, **kwargs)
        self[parameter.name] = parameter

    def value(self, key) -> Any:
        """Return the value for a given parameter.
        """
        return self[key].value

    # pylint: disable=inconsistent-return-statements
    def update_value(self, key, value) -> ParameterValidationError:
        """Update the value of a configuration parameter.
        """
        if key not in self:
            logger.warning(f'Unknow configuration parameter "{key}", skipping...')
            return
        return self[key].set_value(value)

    def update(self, file_path: str) -> None:
        """Update the configuration parameters from a JSON file.
        """
        logger.info(f'Updating configuration from {file_path}...')
        with open(file_path, 'r', encoding=DEFAULT_CHARACTER_ENCODING) as input_file:
            data = json.load(input_file)
        for key, param_dict in data.items():
            self.update_value(key, param_dict['value'])

    def to_json(self) -> str:
        """Encode the configuration into JSON to be written to file.
        """
        data = {key: value.__dict__ for key, value in self.items()}
        return json.dumps(data, indent=4)

    def save(self, file_path) -> None:
        """Dump the configuration to a JSON file.
        """
        logger.info(f'Writing configuration to {file_path}...')
        with open(file_path, 'w', encoding=DEFAULT_CHARACTER_ENCODING) as output_file:
            output_file.write(self.to_json())

    @staticmethod
    def terminal_line(character: str = '-', default_length: int = 50) -> str:
        """Concatenate a series of characters as long as the terminal line.

        Note that we need the try/except block to get this thing working into
        pytest---see https://stackoverflow.com/questions/63345739
        """
        try:
            return character * os.get_terminal_size().columns
        except OSError:
            return character * default_length

    @staticmethod
    def title(text: str) -> str:
        """Pretty-print title.
        """
        line = ConfigurationBase.terminal_line()
        return f'{line}\n{text}\n{line}'

    def __str__(self) -> str:
        """String formatting.
        """
        title = self.title(self.TITLE)
        data = ''.join(f'{param}\n' for param in self.values())
        line = self.terminal_line()
        return f'{title}\n{data}{line}'


class EmptyConfiguration(ConfigurationBase):

    """Empty configuration.
    """

    TITLE = 'Empty configuration'


class SampleConfiguration(ConfigurationBase):

    """Sample configuration.
    """

    TITLE = 'A simple test configuration'
    PARAMETER_SPECS = (
        ('enabled', 'bool', True, 'Enable connection', None, None, {}),
        ('protocol', 'str', 'UDP', 'Communication protocol', None, None,
            dict(choices=('UDP', 'TCP/IP'))),
        ('ip_address', 'str', '127.0.0.1', 'IP address', None, None, {}),
        ('port', 'int', 20004, 'Port', None, None, dict(min=1024, max=65535)),
        ('timeout', 'float', 10., 'Connection timeout', 's', '.3f', dict(min=0.))
    )



# Copyright (C) 2023 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Histogram facilities.
"""

import numpy as np

from baldaquin.plt_ import matplotlib, plt, setup_axes, PlotCard


class InvalidShapeError(RuntimeError):

    """RuntimeError subclass to signal an invalid shape while operating with arrays.
    """

    def __init__(self, expected, actual):
        """Constructor.
        """
        super().__init__(f'Invalid array shape: {expected} expected, got {actual}')


class HistogramBase:

    """Base class for an n-dimensional histogram.

    This interface to histograms is profoundly different for the minimal
    numpy/matplotlib approach, where histogramming methods return bare
    vectors of bin edges and counts.

    Note that this base class is not meant to be instantiated directly, and
    the interfaces to concrete histograms of specific dimensionality are
    defined in the sub-classes.

    Parameters
    ----------
    binning : n-dimensional tuple of arrays
        the bin edges on the different axes.

    labels : n-dimensional tuple of strings
        the text labels for the different axes.
    """

    PLOT_OPTIONS = {}

    def __init__(self, binning, labels):
        """Constructor.
        """
        # Quick check on the binning and label tuples---we need N + 1 labels
        # for an N-dimensional histogram.
        if not len(labels) == len(binning) + 1:
            msg = f'Length mismatch between binning ({len(binning)}) and labels ({len(labels)})'
            raise RuntimeError(msg)
        # The binning is not supposed to change ever, so we turn it into a tuple...
        self.binning = tuple(binning)
        # ...while the labels might conceivably be changed after the fact, hence a list.
        self.labels = list(labels)
        # Initialize all the relevant arrays.
        self._shape = tuple(len(bins) - 1 for bins in self.binning)
        self.content = self._zeros()
        self.entries = self._zeros()
        self._sumw2 = self._zeros()

    def _zeros(self, dtype: type = float):
        """Return an array of zeros of the proper shape for the underlying
        histograms quantities.
        """
        return np.zeros(shape=self._shape, dtype=dtype)

    def reset(self) -> None:
        """Reset the histogram.
        """
        self.content = self._zeros()
        self.entries = self._zeros()
        self._sumw2 = self._zeros()

    def bin_centers(self, axis: int = 0) -> np.array:
        """Return the bin centers for a specific axis.
        """
        return 0.5 * (self.binning[axis][1:] + self.binning[axis][:-1])

    def bin_widths(self, axis: int = 0) -> np.array:
        """Return the bin widths for a specific axis.
        """
        return np.diff(self.binning[axis])

    def errors(self) -> np.array:
        """Return the errors on the bin content.
        """
        return np.sqrt(self._sumw2)

    def set_axis_label(self, axis: int, label: str) -> None:
        """Set the label for a given axis.
        """
        self.labels[axis] = label

    @staticmethod
    def calculate_axis_statistics(bin_centers: np.array, content: np.array) -> dict:
        """Calculate the basic statistics (normalization, mean and rms) for a
        given set of binned data.
        """
        sumw = content.sum()
        if sumw == 0:
            mean = np.nan
            rms = np.nan
        else:
            mean = (bin_centers * content).sum() / sumw
            rms = np.sqrt(((bin_centers - mean)**2. * content).sum() / sumw)
        return {'Sum of weights': sumw, 'Mean':  mean, 'RMS': rms}

    def _check_array_shape(self, data: np.array) -> None:
        """Check the shape of a given array used to update the histogram.
        """
        if data.shape == self._shape:
            raise InvalidShapeError(self._shape, data.shape)

    def set_errors(self, errors: np.array) -> None:
        """Set the proper value for the _sumw2 underlying array, given the
        errors on the bin content.
        """
        self._check_array_shape(errors)
        self._sumw2 = errors**2.

    def fill(self, *values, weights=None):
        """Fill the histogram from unbinned data.

        Note this method is returning the histogram instance, so that the function
        call can be chained.
        """
        values = np.vstack(values).T
        if weights is None:
            content, _ = np.histogramdd(values, bins=self.binning)
            entries = content
            sumw2 = content
        else:
            content, _ = np.histogramdd(values, bins=self.binning, weights=weights)
            entries, _ = np.histogramdd(values, bins=self.binning)
            sumw2, _ = np.histogramdd(values, bins=self.binning, weights=weights**2.)
        self.content += content
        self.entries += entries
        self._sumw2 += sumw2
        return self

    def set_content(self, content: np.array, entries: np.array = None, errors: np.array = None):
        """Set the bin contents programmatically from binned data.

        Note this method is returning the histogram instance, so that the function
        call can be chained.
        """
        self._check_array_shape(errors)
        self.content = content
        if entries is not None:
            self._check_array_shape(entries)
            self.entries = entries
        if errors is not None:
            self.set_errors(errors)
        return self

    @staticmethod
    def bisect(binning: np.array, values: np.array, side: str = 'left') -> np.array:
        """Return the indices corresponding to a given array of values for a
        given binning.
        """
        return np.searchsorted(binning, values, side) - 1

    def find_bin(self, *coords):
        """Find the bin corresponding to a given set of "physical" coordinates
        on the histogram axes.

        This returns a tuple of integer indices that can be used to address
        the histogram content.
        """
        return tuple(self.bisect(binning, value) for binning, value in zip(self.binning, coords))

    def find_bin_value(self, *coords):
        """Find the histogram content corresponding to a given set of "physical"
        coordinates on the histogram axes.
        """
        return self.content[self.find_bin(*coords)]

    def normalization(self, axis: int = None):
        """return the sum of weights in the histogram.
        """
        return self.content.sum(axis)

    def empty_copy(self):
        """Create an empty copy of a histogram.
        """
        return self.__class__(*self.binning, *self.labels)

    def copy(self):
        """Create a full copy of a histogram.
        """
        hist = self.empty_copy()
        hist.set_content(self.content.copy(), self.entries.copy())
        return hist

    def __add__(self, other):
        """Histogram addition.
        """
        hist = self.empty_copy()
        hist.set_content(self.content + other.content, self.entries + other.entries,
                         np.sqrt(self._sumw2 + other._sumw2))
        return hist

    def __sub__(self, other):
        """Histogram subtraction.
        """
        hist = self.empty_copy()
        hist.set_content(self.content - other.content, self.entries + other.entries,
                         np.sqrt(self._sumw2 + other._sumw2))
        return hist

    def __mul__(self, value):
        """Histogram multiplication by a scalar.
        """
        hist = self.empty_copy()
        hist.set_content(self.content * value, self.entries, self.errors() * value)
        return hist

    def __rmul__(self, value):
        """Histogram multiplication by a scalar.
        """
        return self.__mul__(value)

    def _plot(self, **kwargs) -> None:
        """No-op plot() method, to be overloaded by derived classes.
        """
        raise NotImplementedError(f'_plot() not implemented for {self.__class__.__name__}')

    def plot(self, axes=None, **kwargs) -> None:
        """Plot the histogram.
        """
        if axes is None:
            axes = plt.gca()
        for key, value in self.PLOT_OPTIONS.items():
            kwargs.setdefault(key, value)
        self._plot(axes, **kwargs)
        setup_axes(axes, xlabel=self.labels[0], ylabel=self.labels[1])


class Histogram1d(HistogramBase):

    """A one-dimensional histogram.
    """

    PLOT_OPTIONS = dict(lw=1.25, alpha=0.4, histtype='stepfilled')

    def __init__(self, xbinning: np.array, xlabel: str = '', ylabel: str = 'Entries/bin') -> None:
        """Constructor.
        """
        HistogramBase.__init__(self, (xbinning, ), [xlabel, ylabel])

    def current_stats(self) -> dict:
        """Calculate the basic binned statistics for the histogram.
        """
        return self.calculate_axis_statistics(self.bin_centers(0), self.content)

    def stat_box(self, axes=None) -> None:
        """Draw a stat box for the histogram.
        """
        PlotCard(self.current_stats()).draw(axes)

    def _plot(self, axes, **kwargs) -> None:
        """Overloaded make_plot() method.
        """
        # pylint: disable=arguments-differ
        axes.hist(self.bin_centers(0), self.binning[0], weights=self.content, **kwargs)


class Histogram2d(HistogramBase):

    """A two-dimensional histogram.
    """

    PLOT_OPTIONS = dict(cmap=plt.get_cmap('hot'))
    # pylint: disable=invalid-name

    def __init__(self, xbinning, ybinning, xlabel='', ylabel='', zlabel='Entries/bin'):
        """Constructor.
        """
        # pylint: disable=too-many-arguments
        HistogramBase.__init__(self, (xbinning, ybinning), [xlabel, ylabel, zlabel])

    def _plot(self, axes, logz=False, **kwargs):
        """Overloaded make_plot() method.
        """
        # pylint: disable=arguments-differ
        x, y = (v.flatten() for v in np.meshgrid(self.bin_centers(0), self.bin_centers(1)))
        bins = self.binning
        w = self.content.T.flatten()
        if logz:
            # Hack for a deprecated functionality in matplotlib 3.3.0
            # Parameters norm and vmin/vmax should not be used simultaneously
            # If logz is requested, we intercent the bounds when created the norm
            # and refrain from passing vmin/vmax downstream.
            vmin = kwargs.pop('vmin', None)
            vmax = kwargs.pop('vmax', None)
            kwargs.setdefault('norm', matplotlib.colors.LogNorm(vmin, vmax))
        axes.hist2d(x, y, bins, weights=w, **kwargs)
        color_bar = axes.colorbar()
        if self.labels[2] is not None:
            color_bar.set_label(self.labels[2])

    def slice(self, bin_index: int, axis: int = 0):
        """Return a slice of the two-dimensional histogram along the given axis.
        """
        hist = Histogram1d(self.binning[axis], self.labels[axis])
        hist.set_content(self.content[:, bin_index], self.entries[:, bin_index])
        return hist

    def slices(self, axis: int = 0):
        """Return all the slices along a given axis.
        """
        return tuple(self.slice(bin_index, axis) for bin_index in range(self._shape[axis]))

    def hslice(self, bin_index: int):
        """Return the horizontal slice for a given bin.
        """
        return self.slice(bin_index, 0)

    def hslices(self):
        """Return a list of all the horizontal slices.
        """
        return self.slices(0)

    def hbisect(self, y: float):
        """Return the horizontal slice corresponding to a given y value.
        """
        return self.hslice(self.bisect(self.binning[1], y))

    def vslice(self, bin_index):
        """Return the vertical slice for a given bin.
        """
        return self.slice(bin_index, 1)

    def vslices(self):
        """Return a list of all the vertical slices.
        """
        return self.slices(1)

    def vbisect(self, x):
        """Return the vertical slice corresponding to a given y value.
        """
        return self.vslice(self.bisect(self.binning[0], x))



# Copyright (C) 2025 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Test suite for the plasduino project.
"""

from baldaquin import logger
from baldaquin.plasduino.protocol import AnalogReadout, DigitalTransition


def test_protocol():
    """Test the protocol.
    """
    readout = AnalogReadout(0xa2, 1, 1000, 255)
    logger.info(readout)
    logger.info(AnalogReadout.text_header('Something [a. u.]'))
    logger.info(readout.to_text())
    transition = DigitalTransition(0xa1, 1, 1000000)
    logger.info(transition)
    logger.info(DigitalTransition.text_header())
    logger.info(transition.to_text())



# Copyright (C) 2022--2024 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Rudimentary release manager.
"""

from argparse import ArgumentParser
import time


from baldaquin import logger, execute_shell_command, DEFAULT_CHARACTER_ENCODING
from baldaquin import BALDAQUIN_VERSION_FILE_PATH, BALDAQUIN_RELEASE_NOTES_PATH, \
                      BALDAQUIN_TOML_FILE_PATH


INCREMENT_MODES = ('major', 'minor', 'patch')


def write_version_file(version: str, tag_date: str) -> None:
    """Write  the versioning info file with a given version string and build date.
    """
    logger.info(f'Writing versioning info to {BALDAQUIN_VERSION_FILE_PATH}...')
    logger.debug(f'Version: {version}')
    logger.debug(f'Tag date: {tag_date}')
    with open(BALDAQUIN_VERSION_FILE_PATH, 'w',
              encoding=DEFAULT_CHARACTER_ENCODING) as output_file:
        output_file.write(f"VERSION = '{version}'\nTAG_DATE = '{tag_date}'\n")
    logger.info('Done.')


def update_toml_file(old_version, new_version) -> None:
    """Update the pyproject.toml file with the new version.

    We should think about what the proper way to do this is. I see at least three
    avenues to out goal:

    * use the third-party module toml---this would probably be the most elegant
      solution, at the expense of adding one more dependence;
    * use the standard-lib tomllib module (which only support reading toml files)
      and provide a custom implementation for the write part;
    * use the brute force and do everything by hand in text format.

    For the time being I am leaning toward the last option, but we might want to
    revise this in the future.
    """
    # Read (and close) the input file.
    logger.info(f'Reading input file {BALDAQUIN_TOML_FILE_PATH}...')
    with open(BALDAQUIN_TOML_FILE_PATH, 'r', encoding=DEFAULT_CHARACTER_ENCODING) as input_file:
        lines = input_file.readlines()
    # Process the file content in memory---note this includes a check on the
    # version found in the file, and we should run this as early as possible in the
    # release process so that, if anything goes wrong, we do exit before we mess
    # with the repo.
    for i, line in enumerate(lines):
        if line.startswith('version'):
            _, version = line.strip('\n').split(' = ')
            version = version.strip('"')
            if version != old_version:
                raise RuntimeError(f'Unexpected version found in {BALDAQUIN_TOML_FILE_PATH} '
                                   f'({version}) vs ({old_version})')
            logger.debug(f'Version {version} found, target is {new_version}.')
            lines[i] = f'version = "{new_version}"\n'
    # Write the (updated) content to file.
    logger.info(f'Writing output file {BALDAQUIN_TOML_FILE_PATH}...')
    with open(BALDAQUIN_TOML_FILE_PATH, 'w', encoding=DEFAULT_CHARACTER_ENCODING) as output_file:
        output_file.writelines(lines)


def increment_version_file(mode: str, tag_date: str) -> str:
    """Update the __version__.py file.
    """
    logger.info('Updating versioning info file...')
    if mode not in INCREMENT_MODES:
        raise RuntimeError(f'Invalid incerement mode "{mode}"---valid modes are {INCREMENT_MODES}')
    logger.info(f'Reading {BALDAQUIN_VERSION_FILE_PATH}...')
    with open(BALDAQUIN_VERSION_FILE_PATH, 'r',
              encoding=DEFAULT_CHARACTER_ENCODING) as input_file:
        old_version = input_file.readline().split('=')[-1].strip(' \'\n')
    logger.debug(f'Previous version was {old_version}')
    major, minor, patch = (int(item) for item in old_version.split('.'))
    if mode == 'major':
        major += 1
        minor = 0
        patch = 0
    elif mode == 'minor':
        minor += 1
        patch = 0
    elif mode == 'patch':
        patch += 1
    new_version = f'{major}.{minor}.{patch}'
    logger.info(f'Target version is {new_version}')
    update_toml_file(old_version, new_version)
    write_version_file(new_version, tag_date)
    return new_version


def update_release_notes(version: str, tag_date: str) -> None:
    """This is appending the version string and the tag date at the top of the file.
    """
    title = '.. _release_notes:\n\nRelease notes\n=============\n\n'
    logger.info(f'Reading in {BALDAQUIN_RELEASE_NOTES_PATH}...')
    with open(BALDAQUIN_RELEASE_NOTES_PATH, 'r',
              encoding=DEFAULT_CHARACTER_ENCODING) as input_file:
        notes = input_file.read().strip('\n').strip(title)
    logger.info(f'Writing out {BALDAQUIN_RELEASE_NOTES_PATH}...')
    with open(BALDAQUIN_RELEASE_NOTES_PATH, 'w',
              encoding=DEFAULT_CHARACTER_ENCODING) as output_file:
        output_file.writelines(title)
        output_file.writelines(f'\n*baldaquin {version} ({tag_date})*\n\n')
        output_file.writelines(notes)
    logger.info('Done.')


def tag(mode):
    """Tag the actual package.
    """
    tag_date = time.strftime('%a, %d %b %Y %H:%M:%S %z')
    execute_shell_command(['git', 'pull'])
    execute_shell_command(['git', 'status'])
    version = increment_version_file(mode, tag_date)
    update_release_notes(version, tag_date)
    msg = f'Prepare for tag {version}.'
    execute_shell_command(['git', 'commit', '-a', '-m', msg])
    execute_shell_command(['git', 'push'])
    msg = 'Tagging version {version}.'
    execute_shell_command(['git', 'tag', '-a', version, '-m', f'"{msg}"'])
    execute_shell_command(['git', 'push', '--tags'])
    execute_shell_command(['git', 'status'])


if __name__ == '__main__':
    parser = ArgumentParser()
    parser.add_argument('mode', type=str, choices=INCREMENT_MODES, help='Tag increment mode')
    args = parser.parse_args()
    tag(args.mode)



# Copyright (C) 2024 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Plasduino shield information.

See the old plasduino repo for all the gory details:
https://bitbucket.org/lbaldini/plasduino/src/main/shields/
"""


class Lab1:

    """Hardware description for the Lab1 shield, version 2.
    """

    # pylint: disable=too-few-public-methods

    VERSION = 2
    # Analog inputs for the temperature measurements.
    ANALOG_PIN_1 = 0
    ANALOG_PIN_2 = 1
    ANALOG_PINS = (ANALOG_PIN_1, ANALOG_PIN_2)
    SHUNT_RESISTANCE = 10.  # kOhm
    # Digital inputs for the time measurements.
    DIGITAL_PIN_1 = 0
    DIGITAL_PIN_2 = 1
    DIGITAL_PINS = (DIGITAL_PIN_1, DIGITAL_PIN_2)



# Copyright (C) 2022 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Test suite for widgets.py
"""

from loguru import logger

from baldaquin.__qt__ import exec_qapp
from baldaquin.config import SampleConfiguration
from baldaquin.gui import ConfigurationWidget, RunControlCardField, MainWindow, bootstrap_window


def _test_main_window():
    """Create a test window to display a the relevant graphical elements.
    """
    app, window = bootstrap_window(MainWindow)
    config_widget = ConfigurationWidget(SampleConfiguration())
    window.add_tab(config_widget, 'Monitor', 'hub')
    window.add_logger_tab()
    # Interact with the widgets a little bit...
    logger.info('Howdy, partner?')
    window.run_control_card.set(RunControlCardField.UPTIME, 12.)
    window.run_control_card.set(RunControlCardField.TEST_STAND_ID, 1)
    window.run_control_card.set(RunControlCardField.RUN_ID, 313)
    return app, window


if __name__ == '__main__':
    qapp, window = _test_main_window()
    window.show()
    exec_qapp(qapp)



# Copyright (C) 2023 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Test suite for timeline.py
"""

import numpy as np

from baldaquin.hist import Histogram1d
from baldaquin.plt_ import plt


def test_hist1d(size=1000000, mean=10., sigma=2.):
    """A few tests for one-dimensional histograms.
    """
    binning = np.linspace(mean - 5. * sigma, mean + 5. * sigma, 100)
    x = np.random.normal(mean, sigma, size)
    h = Histogram1d(binning, xlabel='x').fill(x)
    # Test that the histogram statistics is providing sensible values.
    # Note that the limits for the test, here, are purely heuristic---we should
    # probably do something more sensible from the statistcal point of view.
    hist_sumw, hist_mean, hist_rms = h.current_stats().values()
    assert abs((hist_sumw - size) / size) < 1.e-5
    assert abs(hist_mean - x.mean()) < 5.e-4
    assert abs(hist_rms - x.std()) < 5.e-3
    plt.figure('Histogram 1d')
    h.plot()
    h.stat_box()


if __name__ == '__main__':
    test_hist1d()
    plt.show()



# Copyright (C) 2022--2025 the baldaquin team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Serial port interface.
"""

from __future__ import annotations

from dataclasses import dataclass
import struct
import time
from typing import Any

import serial
import serial.tools.list_ports
import serial.tools.list_ports_common

from baldaquin import logger


# List of standard baud rates.
STANDARD_BAUD_RATES = serial.Serial.BAUDRATES
DEFAULT_BAUD_RATE = 115200


@dataclass
class DeviceId:

    """Data class to hold the device id information.

    A device id is basically a tuple of two integers, the vendor id (vid) and the
    product id (pid), and the class has exactly these two members.

    Arguments
    ---------
    vid : int
        The vendor id.

    pid : int
        The product id.
    """

    vid: int
    pid: int

    def __eq__(self, other) -> bool:
        """Equality comparison.

        Note we support the comparison between a DeviceId object and a tuple.
        """
        if isinstance(other, tuple):
            return (self.vid, self.pid) == other
        return (self.vid, self.pid) == (other.vid, other.pid)

    @staticmethod
    def _hex(value: int) -> str:
        """Convenience function to format an integer as a hexadecimal string,
        gracefully handling the case when the input is not an integer.
        """
        try:
            return hex(value)
        except TypeError:
            return None

    def __hash__(self):
        """Hash function.
        """
        return hash((self.vid, self.pid))

    def __repr__(self) -> str:
        """String formatting.
        """
        return f'(vid={self._hex(self.vid)}, pid={self._hex(self.pid)})'


@dataclass
class Port:

    """Small data class holding the informatio about a COM port.

    This is a simple wrapper around the serial.tools.list_ports_common.ListPortInfo
    isolating the basic useful functionalities, for the sake of simplicity.

    See https://pyserial.readthedocs.io/en/latest/tools.html#serial.tools.list_ports.ListPortInfo
    for more information.

    Arguments
    ---------
    name : str
        The name of the port (e.g., ``/dev/ttyACM0``).

    device_id : DeviceId
        The device id.

    manufacturer : str, optional
        The manufacturer of the device attached to the port.
    """

    name: str
    device_id: DeviceId
    manufacturer: str = None

    @classmethod
    def from_port_info(cls, port_info: serial.tools.list_ports_common.ListPortInfo) -> 'Port':
        """Create a Port object from a ListPortInfo object.
        """
        device_id = DeviceId(port_info.vid, port_info.pid)
        return cls(port_info.device, device_id, port_info.manufacturer)


def list_com_ports(*device_ids: DeviceId) -> list[Port]:
    """List all the com ports with devices attached, possibly with a filter on
    the device ids we are interested into.

    Arguments
    ---------
    device_ids : DeviceId or vid, pid tuples, optional
        An arbitrary number of device ids to filter the list of ports returned by pyserial.
        This is useful when we are searching for a specific device attached to a
        port; an arduino uno, e.g., might look something like (0x2341, 0x43).

    Returns
    -------
    list of Port objects
        The list of COM ports.
    """
    logger.info('Scanning serial devices...')
    # Populate the initial list of ports.
    ports = [Port.from_port_info(port_info) for port_info in serial.tools.list_ports.comports()]
    for port in ports:
        logger.debug(port)
    logger.info(f'Done, {len(ports)} device(s) found.')
    # If we're not filtering over device ids, we're done.
    if len(device_ids) == 0:
        return ports
    # Otherwise, we filter the list of ports, assuming we have any.
    if len(ports) > 0:
        # If we have a list of tuples, we convert them to DeviceId objects---this
        # will make the printout on the terminal nicer, with the 0x and all that.
        device_ids = list(device_ids)
        for i, entry in enumerate(device_ids):
            if isinstance(entry, tuple):
                device_ids[i] = DeviceId(*entry)
        logger.info(f'Filtering port list for specific devices: {device_ids}...')
        # Do the actual filtering.
        ports = [port for port in ports if port.device_id in device_ids]
        logger.info(f'Done, {len(ports)} device(s) remaining.')
    for port in ports:
        logger.debug(port)
    return ports


class SerialInterface(serial.Serial):

    """Small wrapper around the serial.Serial class.
    """

    # pylint: disable=too-many-ancestors

    def setup(self, port: str, baudrate: int = DEFAULT_BAUD_RATE, timeout: float = None) -> None:
        """Setup the serial connection.

        Arguments
        ---------
        port : str
            The name of the port to connect to (e.g., ``/dev/ttyACM0``).

        baudrate : int
            The baud rate.

            Verbatim from the pyserial documentation: the parameter baudrate can
            be one of the standard values: 50, 75, 110, 134, 150, 200, 300, 600,
            1200, 1800, 2400, 4800, 9600, 19200, 38400, 57600, 115200. These are
            well supported on all platforms.

            Standard values above 115200, such as: 230400, 460800, 500000, 576000,
            921600, 1000000, 1152000, 1500000, 2000000, 2500000, 3000000, 3500000,
            4000000 also work on many platforms and devices.

            Non-standard values are also supported on some platforms (GNU/Linux,
            MAC OSX >= Tiger, Windows). Though, even on these platforms some serial
            ports may reject non-standard values.

        timeout : float
            The timeout in seconds.

            Verbatim from the pyserial documentation: possible values for the parameter
            timeout which controls the behavior of read():

            * ``timeout = None``: wait forever / until requested number of bytes
              are received

            * ``timeout = 0``: non-blocking mode, return immediately in any case,
              returning zero or more, up to the requested number of bytes

            * ``timeout = x``: set timeout to x seconds (float allowed) returns
              immediately when the requested number of bytes are available,
              otherwise wait until the timeout expires and return all bytes that
              were received until then.
        """
        logger.debug(f'Configuring serial connection (port = {port}, '
                     f'baudarate = {baudrate}, timeout = {timeout})...')
        self.port = port
        self.baudrate = baudrate
        self.timeout = timeout

    def connect(self, port: str, baudrate: int = 115200, timeout: float = None) -> None:
        """Connect to the serial port.

        Arguments
        ---------
        port : str
            The name of the serial port (e.g., ``/dev/ttyACM0``).

        baudrate : int
            The baud rate.

        timeout : float, optional
            The timeout in seconds.
        """
        self.setup(port, baudrate, timeout)
        logger.info(f'Opening serial connection to port {self.port}...')
        self.open()

    def disconnect(self):
        """Disconnect from the serial port.
        """
        logger.info(f'Closing serial connection to port {self.port}...')
        self.close()

    def pulse_dtr(self, pulse_length: float = 0.5) -> None:
        """Pulse the DTR line for a given amount of time.

        This asserts the DTR line, waits for a specific amount of time, and then
        deasserts the line.

        Arguments
        ---------
        pulse_length : float
            The duration (in seconds) for the DTR line signal to be asserted.
        """
        logger.info(f'Pulsing the DTR line for {pulse_length} s...')
        self.dtr = 1
        time.sleep(pulse_length)
        self.dtr = 0

    def read_and_unpack(self, fmt: str) -> Any:
        """Read a given number of bytes from the serial port and unpack them.

        Note that the number of bytes to be read from the serial port is automatically
        calculated from the format string.
        See https://docs.python.org/3/library/struct.html for all the details about
        format strings and byte ordering.

        Arguments
        ---------
        fmt : str
            The format string for the packet to be read from the seria port.

        Returns
        -------
        any
            Returns the proper Python object for the format string at hand.

        Example
        -------
        >>> s = SerialInterface(port)
        >>> val = s.read_and_unpack('B') # Single byte (val is int)
        >>> val = s.read_and_unpack('>L') # Big-endian unsigned long (val is also int)
        """
        data = self.read(struct.calcsize(fmt))
        try:
            return struct.unpack(fmt, data)[0]
        except struct.error as exception:
            logger.error(f'Could not unpack {data} with format "{fmt}".')
            raise exception

    def pack_and_write(self, value: Any, fmt: str) -> int:
        """ Pack a given value into a proper bytes object and write the latter
        to the serial port.

        Arguments
        ---------
        value : any
            The value to be written to the serial port.

        fmt : str
            The format string to pack the value with.

        Returns
        -------
        int
            The number of bytes written to the serial port.
        """
        return self.write(struct.pack(fmt, value))
